/***********************************************************************/
/**
$Header: /Repository/drv/linux/drv/hpidspcd.c,v 1.4 2011/04/12 11:22:14 as-ewb Exp $

\file
Functions for reading DSP code using
hotplug firmware loader from individual dsp code files
*/
/***********************************************************************/
#define SOURCEFILE_NAME "hpidspcd.c"
#include "hpidspcd.h"
#include "hpidebug.h"
#include "linux/pci.h"

struct dsp_code_private {
	/**  Firmware descriptor */
	const struct firmware *firmware;
	struct pci_dev *dev;
};

#define HPI_VER_DECIMAL ((int)(HPI_VER_MAJOR(HPI_VER) * 10000 + \
            HPI_VER_MINOR(HPI_VER) * 100 + HPI_VER_RELEASE(HPI_VER)))

/*-------------------------------------------------------------------*/
short HpiDspCode_Open(uint32_t adapter, void *os_data,
		      struct dsp_code *dsp_code, uint32_t *os_error_code)
{
	const struct firmware *firmware;
	struct pci_dev *dev = os_data;
	struct code_header header;
	char fw_name[20];
	int err;

	sprintf(fw_name, "asihpi/dsp%04x.bin", adapter);

	err = request_firmware(&firmware, fw_name, &dev->dev);

	if (err || !firmware) {
		dev_printk(KERN_ERR,  &dev->dev,
		"%d, request_firmware failed for  %s\n",
			       err, fw_name);
		goto error1;
	}
	if (firmware->size < sizeof(header)) {
		dev_printk(KERN_ERR,  &dev->dev,
		"Header size too small %s\n", fw_name);
		goto error2;
	}
	memcpy(&header, firmware->data, sizeof(header));

	if ((header.type != 0x45444F43) || /* "CODE" */
	    (header.adapter != adapter) ||
	    (header.size != firmware->size)) {
		dev_printk(KERN_ERR,  &dev->dev,
		    "Invalid firmware file\n");
		goto error2;
	}

	if ((header.version / 100 & ~1) != (HPI_VER_DECIMAL / 100 & ~1)) {
		dev_printk(KERN_ERR, &dev->dev,
		    "Incompatible firmware version "
		    "DSP image %d != Driver %d\n",
		    header.version, HPI_VER_DECIMAL);
		goto error2;
	}

	if (header.version != HPI_VER_DECIMAL) {
		dev_printk(KERN_WARNING, &dev->dev,
		    "Firmware: release version mismatch  DSP image %d != Driver %d\n",
		    header.version, HPI_VER_DECIMAL);
	}

	HPI_DEBUG_LOG(DEBUG, "Dsp code %s opened\n", fw_name);
	dsp_code->pvt = kmalloc(sizeof(*dsp_code->pvt), GFP_KERNEL);
	if (!dsp_code->pvt)
		return HPI_ERROR_MEMORY_ALLOC;

	dsp_code->pvt->dev = dev;
	dsp_code->pvt->firmware = firmware;
	dsp_code->header = header;
	dsp_code->block_length = header.size / sizeof(uint32_t);
	dsp_code->word_count = sizeof(header) / sizeof(uint32_t);
	return 0;

error2:
	release_firmware(firmware);
error1:
	dsp_code->block_length = 0;
	return HPI_ERROR_DSP_FILE_NOT_FOUND;
}

/*-------------------------------------------------------------------*/
void HpiDspCode_Close(struct dsp_code *dsp_code)
{
	if (dsp_code->pvt->firmware) {
		HPI_DEBUG_LOG(DEBUG, "Dsp code closed\n");
		release_firmware(dsp_code->pvt->firmware);
		dsp_code->pvt->firmware = NULL;
	}
	kfree(dsp_code->pvt);
}

/*-------------------------------------------------------------------*/
void HpiDspCode_Rewind(struct dsp_code *dsp_code)
{
	/* Go back to start of  data, after header */
	dsp_code->word_count = sizeof(struct code_header) / sizeof(uint32_t);
}

/*-------------------------------------------------------------------*/
short HpiDspCode_ReadWord(struct dsp_code *dsp_code, uint32_t *pdwWord)
{
	if (dsp_code->word_count + 1 > dsp_code->block_length)
		return HPI_ERROR_DSP_FILE_FORMAT;

	*pdwWord =
	    ((uint32_t *)(dsp_code->pvt->firmware->data))[dsp_code->word_count];
	dsp_code->word_count++;
	return 0;
}

/*-------------------------------------------------------------------*/
short HpiDspCode_ReadBlock(size_t nWordsRequested,
			   struct dsp_code *dsp_code, uint32_t **ppdwBlock)
{
	if (dsp_code->word_count + nWordsRequested >
	    dsp_code->block_length)
		return HPI_ERROR_DSP_FILE_FORMAT;

	*ppdwBlock =
	    ((uint32_t *)(dsp_code->pvt->firmware->data)) + dsp_code->word_count;
	dsp_code->word_count += nWordsRequested;
	return 0;
}

// $Log: hpidspcd.c,v $
// Revision 1.4  2011/04/12 11:22:14  as-ewb
// Clean hpidspcode api
//
// Revision 1.3  2011/04/06 04:50:52  as-ewb
// Move towards os independent dsp code loader API
//
// Revision 1.1  2011/04/06 03:32:23  as-ewb
// linux driver specific dsp code loader in linux driver directory
//
