/******************************************************************************

    AudioScience HPI driver
    Copyright (C) 1997-2011  AudioScience Inc. <support@audioscience.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as
    published by the Free Software Foundation;

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


Extended Message Function With Response Caching

(C) Copyright AudioScience Inc. 2002
*****************************************************************************/
#define SOURCEFILE_NAME "hpimsgx.c"
#include "hpi_internal.h"
#include "hpimsginit.h"
#include "hpicmn.h"
#include "hpimsgx.h"
#include "hpidebug.h"

static struct pci_device_id asihpi_pci_tbl[] = {
#include "hpipcida.h"
};

#ifdef HPI_LOCKING
static struct hpios_spinlock msgxLock;
#endif

static HPI_HandlerFunc *hpi_entry_points[HPI_MAX_ADAPTERS];

static HPI_HandlerFunc *HPI_LookupEntryPointFunction(
	const struct hpi_pci *PciInfo
)
{

	int i;

	for (i = 0; asihpi_pci_tbl[i].vendor != 0; i++) {
		if (asihpi_pci_tbl[i].vendor != PCI_ANY_ID
			&& asihpi_pci_tbl[i].vendor != HPIOS_PCI_VENDOR(PciInfo))
			continue;
		if (asihpi_pci_tbl[i].device != PCI_ANY_ID
			&& asihpi_pci_tbl[i].device != HPIOS_PCI_DEVICE(PciInfo))
			continue;
		if (asihpi_pci_tbl[i].subvendor != PCI_ANY_ID
			&& asihpi_pci_tbl[i].subvendor !=
				HPIOS_PCI_SUBSYS_VENDOR(PciInfo))
			continue;
		if (asihpi_pci_tbl[i].subdevice != PCI_ANY_ID
			&& asihpi_pci_tbl[i].subdevice !=
				HPIOS_PCI_SUBSYS_DEVICE(PciInfo))
			continue;

		/* HPI_DEBUG_LOG2(DEBUG, " %x,%lx\n", i,
			asihpi_pci_tbl[i].driver_data); */
		return (HPI_HandlerFunc *) asihpi_pci_tbl[i].driver_data;
	}

	return NULL;
}

static inline void HwEntryPoint(
	struct hpi_message *phm,
	struct hpi_response *phr
)
{
	if ((phm->wAdapterIndex < HPI_MAX_ADAPTERS) &&
		hpi_entry_points[phm->wAdapterIndex])
		hpi_entry_points[phm->wAdapterIndex](phm, phr);
	else
		HPI_InitResponse(phr, phm->wObject, phm->wFunction,
			HPI_ERROR_PROCESSING_MESSAGE);
}

static void AdapterOpen(
	struct hpi_message *phm,
	struct hpi_response *phr
);
static void AdapterClose(
	struct hpi_message *phm,
	struct hpi_response *phr
);

static void MixerOpen(
	struct hpi_message *phm,
	struct hpi_response *phr
);
static void MixerClose(
	struct hpi_message *phm,
	struct hpi_response *phr
);

static void OutStreamOpen(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
);
static void OutStreamClose(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
);
static void InStreamOpen(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
);
static void InStreamClose(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
);

static void HPIMSGX_Reset(uint16_t wAdapterIndex);

static uint16_t HPIMSGX_Init(
	struct hpi_message *phm,
	struct hpi_response *phr
);
static void HPIMSGX_Cleanup(
	uint16_t wAdapterIndex,
	void *hOwner
);

#ifndef DISABLE_PRAGMA_PACK1
#pragma pack(push, 1)
#endif

struct hpi_subsys_response {
	struct hpi_response_header h;
	struct hpi_subsys_res s;
};

struct hpi_adapter_response {
	struct hpi_response_header h;
	struct hpi_adapter_res a;
};

struct hpi_mixer_response {
	struct hpi_response_header h;
	struct hpi_mixer_res m;
};

struct hpi_stream_response {
	struct hpi_response_header h;
	struct hpi_stream_res d;
};

struct adapter_info {
	uint16_t wType;
	uint16_t wNumInStreams;
	uint16_t wNumOutStreams;
};

struct asi_open_state {
	int nOpenFlag;
	void *hOwner;
};

#ifndef DISABLE_PRAGMA_PACK1
#pragma pack(pop)
#endif

// Globals
static struct hpi_adapter_response aRESP_HPI_ADAPTER_OPEN[HPI_MAX_ADAPTERS];

static struct hpi_stream_response
	aRESP_HPI_OSTREAM_OPEN[HPI_MAX_ADAPTERS][HPI_MAX_STREAMS];

static struct hpi_stream_response
	aRESP_HPI_ISTREAM_OPEN[HPI_MAX_ADAPTERS][HPI_MAX_STREAMS];

static struct hpi_mixer_response aRESP_HPI_MIXER_OPEN[HPI_MAX_ADAPTERS];

static struct adapter_info aADAPTER_INFO[HPI_MAX_ADAPTERS];

// use these to keep track of opens from user mode apps/DLLs
static struct asi_open_state
	aOStreamUserOpen[HPI_MAX_ADAPTERS][HPI_MAX_STREAMS];

static struct asi_open_state
	aIStreamUserOpen[HPI_MAX_ADAPTERS][HPI_MAX_STREAMS];

static void SubSysMessage(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
)
{
	if (phm->wAdapterIndex != HPI_ADAPTER_INDEX_INVALID)
		HPI_DEBUG_LOG2(WARNING, "Suspicious adapter index %d in subsys message 0x%x.\n",
			phm->wAdapterIndex, phm->wFunction);

	switch (phm->wFunction) {
	case HPI_SUBSYS_GET_VERSION:
		HPI_InitResponse(phr, HPI_OBJ_SUBSYSTEM,
			HPI_SUBSYS_GET_VERSION, 0);
		phr->u.s.dwVersion = HPI_VER >> 8;	// return major.minor
		phr->u.s.dwData = HPI_VER;	// return major.minor.release
		break;
	case HPI_SUBSYS_OPEN:
		//do not propagate the message down the chain
		HPI_InitResponse(phr, HPI_OBJ_SUBSYSTEM, HPI_SUBSYS_OPEN, 0);
		break;
	case HPI_SUBSYS_CLOSE:
		//do not propagate the message down the chain
		HPI_InitResponse(phr, HPI_OBJ_SUBSYSTEM, HPI_SUBSYS_CLOSE, 0);
		HPIMSGX_Cleanup(HPIMSGX_ALLADAPTERS, hOwner);
		break;
	case HPI_SUBSYS_DRIVER_LOAD:
		// Initialize this module's internal state
		HpiOs_Msgxlock_Init(&msgxLock);
		memset(&hpi_entry_points, 0, sizeof(hpi_entry_points));
		HpiOs_LockedMem_Init();
		// Init subsys_findadapters response to no-adapters
		HPIMSGX_Reset(HPIMSGX_ALLADAPTERS);
		HPI_InitResponse(phr, HPI_OBJ_SUBSYSTEM,
			HPI_SUBSYS_DRIVER_LOAD, 0);
		/* individual HPIs dont implement driver load */
		HPI_COMMON(phm, phr);
		break;
	case HPI_SUBSYS_DRIVER_UNLOAD:
		HPI_COMMON(phm, phr);
		HPIMSGX_Cleanup(HPIMSGX_ALLADAPTERS, hOwner);
		HpiOs_LockedMem_FreeAll();
		HPI_InitResponse(phr, HPI_OBJ_SUBSYSTEM,
			HPI_SUBSYS_DRIVER_UNLOAD, 0);
		return;

	case HPI_SUBSYS_GET_NUM_ADAPTERS:
	case HPI_SUBSYS_GET_ADAPTER:
		HPI_COMMON(phm, phr);
		break;

	case HPI_SUBSYS_CREATE_ADAPTER:
		HPIMSGX_Init(phm, phr);
		break;

	default:
		// Must explicitly handle every subsys message in this switch
		HPI_InitResponse(phr, HPI_OBJ_SUBSYSTEM, phm->wFunction,
						HPI_ERROR_INVALID_FUNC);
		break;
	}
}

static void AdapterMessage(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
)
{
	switch (phm->wFunction) {
	case HPI_ADAPTER_OPEN:
		AdapterOpen(phm, phr);
		break;
	case HPI_ADAPTER_CLOSE:
		AdapterClose(phm, phr);
		break;
	case HPI_ADAPTER_DELETE:
		HPIMSGX_Cleanup(phm->wAdapterIndex, hOwner);
		{
			struct hpi_message hm;
			struct hpi_response hr;
			HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_ADAPTER,
				HPI_ADAPTER_CLOSE);
			hm.wAdapterIndex = phm->wAdapterIndex;
			HwEntryPoint(&hm, &hr);
		}
		HwEntryPoint(phm, phr);
		break;

	default:
		HwEntryPoint(phm, phr);
		break;
	}
}

static void MixerMessage(
	struct hpi_message *phm,
	struct hpi_response *phr
)
{
	switch (phm->wFunction) {
	case HPI_MIXER_OPEN:
		MixerOpen(phm, phr);
		break;
	case HPI_MIXER_CLOSE:
		MixerClose(phm, phr);
		break;
	default:
		HwEntryPoint(phm, phr);
		break;
	}
}

static void OStreamMessage(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
)
{
	if (phm->wObjIndex >=
		aADAPTER_INFO[phm->wAdapterIndex].wNumOutStreams) {
		HPI_InitResponse(phr, HPI_OBJ_OSTREAM, phm->wFunction,
			HPI_ERROR_INVALID_OBJ_INDEX);
		return;
	}

	switch (phm->wFunction) {
	case HPI_OSTREAM_OPEN:
		OutStreamOpen(phm, phr, hOwner);
		break;
	case HPI_OSTREAM_CLOSE:
		OutStreamClose(phm, phr, hOwner);
		break;
	default:
		HwEntryPoint(phm, phr);
		break;
	}
}

static void IStreamMessage(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
)
{
	if (phm->wObjIndex >=
		aADAPTER_INFO[phm->wAdapterIndex].wNumInStreams) {
		HPI_InitResponse(phr, HPI_OBJ_ISTREAM, phm->wFunction,
			HPI_ERROR_INVALID_OBJ_INDEX);
		return;
	}

	switch (phm->wFunction) {
	case HPI_ISTREAM_OPEN:
		InStreamOpen(phm, phr, hOwner);
		break;
	case HPI_ISTREAM_CLOSE:
		InStreamClose(phm, phr, hOwner);
		break;
	default:
		HwEntryPoint(phm, phr);
		break;
	}
}

/* NOTE: HPI_Message() must be defined in the driver as a wrapper for
 * HPI_MessageEx so that functions in hpifunc.c compile.
 */
void HPI_MessageEx(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
)
{
	HPI_DEBUG_MESSAGE(DEBUG, phm);

	if (phm->wType != HPI_TYPE_REQUEST) {
		HPI_InitResponse(phr, phm->wObject, phm->wFunction,
			HPI_ERROR_INVALID_TYPE);
		return;
	}

	if (phm->wAdapterIndex >= HPI_MAX_ADAPTERS &&
		phm->wAdapterIndex != HPIMSGX_ALLADAPTERS) {
		HPI_InitResponse(phr, phm->wObject, phm->wFunction,
			HPI_ERROR_BAD_ADAPTER_NUMBER);
		return;
	}

	switch (phm->wObject) {
	case HPI_OBJ_SUBSYSTEM:
		SubSysMessage(phm, phr, hOwner);
		break;

	case HPI_OBJ_ADAPTER:
		AdapterMessage(phm, phr, hOwner);
		break;

	case HPI_OBJ_MIXER:
		MixerMessage(phm, phr);
		break;

	case HPI_OBJ_OSTREAM:
		OStreamMessage(phm, phr, hOwner);
		break;

	case HPI_OBJ_ISTREAM:
		IStreamMessage(phm, phr, hOwner);
		break;

	default:
		HwEntryPoint(phm, phr);
		break;
	}
	HPI_DEBUG_RESPONSE(phr);

#ifndef HPI_OS_LINUX_KERNEL
	if ((phr->wError == HPI_ERROR_DSP_COMMUNICATION) ||
		(phr->wError == HPI_ERROR_DSP_BOOTLOAD))
	{
		void * ep = NULL;
		char * ep_name;
		char * err_name;

		HPI_DEBUG_MESSAGE(ERROR, phm);

		if (phm->wAdapterIndex < HPI_MAX_ADAPTERS)
			ep = hpi_entry_points[phm->wAdapterIndex];

		hpidebug_backend_error(ep, phr->wSpecificError, &ep_name, &err_name);

		HPI_DEBUG_LOG4(ERROR,
				"Adap=%d %s ERROR# %d, %s\n",
				phm->wAdapterIndex,
				ep_name,
				phr->wError,
				err_name);

		if (hpiDebugLevel >= HPI_DEBUG_LEVEL_VERBOSE)
			hpi_debug_data((uint16_t *)phm,
				sizeof(*phm)/sizeof(uint16_t));
	}
#endif
}

static void AdapterOpen(
	struct hpi_message *phm,
	struct hpi_response *phr
)
{
	HPI_DEBUG_LOG0(VERBOSE, "AdapterOpen\n");
	memcpy(phr, &aRESP_HPI_ADAPTER_OPEN[phm->wAdapterIndex],
		sizeof(aRESP_HPI_ADAPTER_OPEN[0]));
}

static void AdapterClose(
	struct hpi_message *phm,
	struct hpi_response *phr
)
{
	HPI_DEBUG_LOG0(VERBOSE, "AdapterClose\n");
	HPI_InitResponse(phr, HPI_OBJ_ADAPTER, HPI_ADAPTER_CLOSE, 0);
}

static void MixerOpen(
	struct hpi_message *phm,
	struct hpi_response *phr
)
{
	memcpy(phr, &aRESP_HPI_MIXER_OPEN[phm->wAdapterIndex],
		sizeof(aRESP_HPI_MIXER_OPEN[0]));
}

static void MixerClose(
	struct hpi_message *phm,
	struct hpi_response *phr
)
{
	HPI_InitResponse(phr, HPI_OBJ_MIXER, HPI_MIXER_CLOSE, 0);
}

static void InStreamOpen(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
)
{

	struct hpi_message hm;
	struct hpi_response hr;

	HPI_InitResponse(phr, HPI_OBJ_ISTREAM, HPI_ISTREAM_OPEN, 0);

	HpiOs_Msgxlock_Lock(&msgxLock);

	if (aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].
				nOpenFlag)
		phr->wError = HPI_ERROR_OBJ_ALREADY_OPEN;
	else if (aRESP_HPI_ISTREAM_OPEN[phm->wAdapterIndex]
				[phm->wObjIndex].h.wError)
		memcpy(phr,
			&aRESP_HPI_ISTREAM_OPEN[phm->wAdapterIndex][phm->wObjIndex],
			sizeof(aRESP_HPI_ISTREAM_OPEN[0][0]));
	else {
		aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].
			nOpenFlag = 1;
		HpiOs_Msgxlock_Unlock(&msgxLock);

		// issue a reset
		HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_ISTREAM, HPI_ISTREAM_RESET);
		hm.wAdapterIndex = phm->wAdapterIndex;
		hm.wObjIndex = phm->wObjIndex;
		HwEntryPoint(&hm, &hr);

		HpiOs_Msgxlock_Lock(&msgxLock);
		if (hr.wError) {
			aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].nOpenFlag = 0;
			phr->wError = hr.wError;
		} else {
			aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].nOpenFlag = 1;
			aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].hOwner = hOwner;
			memcpy(phr,
				&aRESP_HPI_ISTREAM_OPEN[phm->
					wAdapterIndex][phm->wObjIndex],
				sizeof(aRESP_HPI_ISTREAM_OPEN[0][0]));
		}
	}
	HpiOs_Msgxlock_Unlock(&msgxLock);
}

static void InStreamClose(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
)
{

	struct hpi_message hm;
	struct hpi_response hr;

	HPI_InitResponse(phr, HPI_OBJ_ISTREAM, HPI_ISTREAM_CLOSE, 0);

	HpiOs_Msgxlock_Lock(&msgxLock);
	if (hOwner ==
		aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].
		hOwner) {
		/* HPI_DEBUG_LOG3(INFO,"closing adapter %d "
		   "instream %d owned by %p\n",
		   phm->wAdapterIndex, phm->wObjIndex, hOwner); */
		aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].
			hOwner = NULL;
		HpiOs_Msgxlock_Unlock(&msgxLock);
		// issue a reset
		HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_ISTREAM, HPI_ISTREAM_RESET);
		hm.wAdapterIndex = phm->wAdapterIndex;
		hm.wObjIndex = phm->wObjIndex;
		HwEntryPoint(&hm, &hr);
		HpiOs_Msgxlock_Lock(&msgxLock);
		if (hr.wError) {
			aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].hOwner = hOwner;
			phr->wError = hr.wError;
		} else {
			aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].nOpenFlag = 0;
			aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].hOwner = NULL;
		}
	} else {
		HPI_DEBUG_LOG4(WARNING,
			"%p trying to close %d instream %d owned by %p\n",
			hOwner, phm->wAdapterIndex,
			phm->wObjIndex,
			aIStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].hOwner);
		phr->wError = HPI_ERROR_OBJ_NOT_OPEN;
	}
	HpiOs_Msgxlock_Unlock(&msgxLock);
}

static void OutStreamOpen(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
)
{

	struct hpi_message hm;
	struct hpi_response hr;

	HPI_InitResponse(phr, HPI_OBJ_OSTREAM, HPI_OSTREAM_OPEN, 0);

	HpiOs_Msgxlock_Lock(&msgxLock);

	if (aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].
		nOpenFlag)
		phr->wError = HPI_ERROR_OBJ_ALREADY_OPEN;
	else if (aRESP_HPI_OSTREAM_OPEN[phm->wAdapterIndex]
		[phm->wObjIndex].h.wError)
		memcpy(phr,
			&aRESP_HPI_OSTREAM_OPEN[phm->wAdapterIndex][phm->wObjIndex],
			sizeof(aRESP_HPI_OSTREAM_OPEN[0][0]));
	else {
		aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].
			nOpenFlag = 1;
		HpiOs_Msgxlock_Unlock(&msgxLock);

		// issue a reset
		HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_OSTREAM, HPI_OSTREAM_RESET);
		hm.wAdapterIndex = phm->wAdapterIndex;
		hm.wObjIndex = phm->wObjIndex;
		HwEntryPoint(&hm, &hr);

		HpiOs_Msgxlock_Lock(&msgxLock);
		if (hr.wError) {
			aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].nOpenFlag = 0;
			phr->wError = hr.wError;
		} else {
			aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].nOpenFlag = 1;
			aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].hOwner = hOwner;
			memcpy(phr,
				&aRESP_HPI_OSTREAM_OPEN[phm->
					wAdapterIndex][phm->wObjIndex],
				sizeof(aRESP_HPI_OSTREAM_OPEN[0][0]));
		}
	}
	HpiOs_Msgxlock_Unlock(&msgxLock);
}

static void OutStreamClose(
	struct hpi_message *phm,
	struct hpi_response *phr,
	void *hOwner
)
{

	struct hpi_message hm;
	struct hpi_response hr;

	HPI_InitResponse(phr, HPI_OBJ_OSTREAM, HPI_OSTREAM_CLOSE, 0);

	HpiOs_Msgxlock_Lock(&msgxLock);

	if (hOwner ==
		aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].
		hOwner) {
		/* HPI_DEBUG_LOG3(INFO,"closing adapter %d "
		   "outstream %d owned by %p\n",
		   phm->wAdapterIndex, phm->wObjIndex, hOwner); */
		aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].
			hOwner = NULL;
		HpiOs_Msgxlock_Unlock(&msgxLock);
		// issue a reset
		HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_OSTREAM, HPI_OSTREAM_RESET);
		hm.wAdapterIndex = phm->wAdapterIndex;
		hm.wObjIndex = phm->wObjIndex;
		HwEntryPoint(&hm, &hr);
		HpiOs_Msgxlock_Lock(&msgxLock);
		if (hr.wError) {
			aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].hOwner = hOwner;
			phr->wError = hr.wError;
		} else {
			aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].nOpenFlag = 0;
			aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].hOwner = NULL;
		}
	} else {
		HPI_DEBUG_LOG4(WARNING,
			"%p trying to close %d outstream %d owned by %p\n",
			hOwner, phm->wAdapterIndex,
			phm->wObjIndex,
			aOStreamUserOpen[phm->wAdapterIndex][phm->wObjIndex].hOwner);
		phr->wError = HPI_ERROR_OBJ_NOT_OPEN;
	}
	HpiOs_Msgxlock_Unlock(&msgxLock);
}

static uint16_t AdapterPrepare(uint16_t wAdapter)
{
	struct hpi_message hm;
	struct hpi_response hr;

	// Open the adapter and streams
	uint16_t i;

	// call to HPI_ADAPTER_OPEN
	HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_ADAPTER, HPI_ADAPTER_OPEN);
	hm.wAdapterIndex = wAdapter;
	HwEntryPoint(&hm, &hr);
	memcpy(&aRESP_HPI_ADAPTER_OPEN[wAdapter], &hr,
		sizeof(aRESP_HPI_ADAPTER_OPEN[0]));
	if (hr.wError)
		return hr.wError;

	// call to HPI_ADAPTER_GET_INFO
	HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_ADAPTER, HPI_ADAPTER_GET_INFO);
	hm.wAdapterIndex = wAdapter;
	HwEntryPoint(&hm, &hr);
	if (hr.wError)
		return hr.wError;

	aADAPTER_INFO[wAdapter].wNumOutStreams = hr.u.ax.info.wNumOStreams;
	aADAPTER_INFO[wAdapter].wNumInStreams = hr.u.ax.info.wNumIStreams;
	aADAPTER_INFO[wAdapter].wType = hr.u.ax.info.wAdapterType;

	// call to HPI_OSTREAM_OPEN
	for (i = 0; i < aADAPTER_INFO[wAdapter].wNumOutStreams; i++) {
		HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_OSTREAM, HPI_OSTREAM_OPEN);
		hm.wAdapterIndex = wAdapter;
		hm.wObjIndex = i;
		HwEntryPoint(&hm, &hr);
		memcpy(&aRESP_HPI_OSTREAM_OPEN[wAdapter][i], &hr,
			sizeof(aRESP_HPI_OSTREAM_OPEN[0][0]));
		aOStreamUserOpen[wAdapter][i].nOpenFlag = 0;
		aOStreamUserOpen[wAdapter][i].hOwner = NULL;
	}

	// call to HPI_ISTREAM_OPEN
	for (i = 0; i < aADAPTER_INFO[wAdapter].wNumInStreams; i++) {
		HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_ISTREAM, HPI_ISTREAM_OPEN);
		hm.wAdapterIndex = wAdapter;
		hm.wObjIndex = i;
		HwEntryPoint(&hm, &hr);
		memcpy(&aRESP_HPI_ISTREAM_OPEN[wAdapter][i], &hr,
			sizeof(aRESP_HPI_ISTREAM_OPEN[0][0]));
		aIStreamUserOpen[wAdapter][i].nOpenFlag = 0;
		aIStreamUserOpen[wAdapter][i].hOwner = NULL;
	}

	// call to HPI_MIXER_OPEN
	HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_MIXER, HPI_MIXER_OPEN);
	hm.wAdapterIndex = wAdapter;
	HwEntryPoint(&hm, &hr);
	memcpy(&aRESP_HPI_MIXER_OPEN[wAdapter], &hr,
		sizeof(aRESP_HPI_MIXER_OPEN[0]));

	return 0;
}

static void HPIMSGX_Reset(uint16_t wAdapterIndex)
{
	int i;
	uint16_t wAdapter;
	struct hpi_response hr;

	if (wAdapterIndex == HPIMSGX_ALLADAPTERS) {
		for (wAdapter = 0; wAdapter < HPI_MAX_ADAPTERS; wAdapter++) {

			HPI_InitResponse(&hr, HPI_OBJ_ADAPTER,
				HPI_ADAPTER_OPEN, HPI_ERROR_BAD_ADAPTER);
			memcpy(&aRESP_HPI_ADAPTER_OPEN[wAdapter], &hr,
				sizeof(aRESP_HPI_ADAPTER_OPEN[wAdapter]));

			HPI_InitResponse(&hr, HPI_OBJ_MIXER, HPI_MIXER_OPEN,
				HPI_ERROR_INVALID_OBJ);
			memcpy(&aRESP_HPI_MIXER_OPEN[wAdapter], &hr,
				sizeof(aRESP_HPI_MIXER_OPEN[wAdapter]));

			for (i = 0; i < HPI_MAX_STREAMS; i++) {
				HPI_InitResponse(&hr, HPI_OBJ_OSTREAM,
					HPI_OSTREAM_OPEN,
					HPI_ERROR_INVALID_OBJ);
				memcpy(&aRESP_HPI_OSTREAM_OPEN[wAdapter][i],
					&hr,
					sizeof(aRESP_HPI_OSTREAM_OPEN
						[wAdapter]
						[i]));
				HPI_InitResponse(&hr, HPI_OBJ_ISTREAM,
					HPI_ISTREAM_OPEN,
					HPI_ERROR_INVALID_OBJ);
				memcpy(&aRESP_HPI_ISTREAM_OPEN[wAdapter][i],
					&hr,
					sizeof(aRESP_HPI_ISTREAM_OPEN
						[wAdapter]
						[i]));
			}
		}
	} else if (wAdapterIndex < HPI_MAX_ADAPTERS) {
		aRESP_HPI_ADAPTER_OPEN[wAdapterIndex].h.wError =
			HPI_ERROR_BAD_ADAPTER;
		aRESP_HPI_MIXER_OPEN[wAdapterIndex].h.wError =
			HPI_ERROR_INVALID_OBJ;
		for (i = 0; i < HPI_MAX_STREAMS; i++) {
			aRESP_HPI_OSTREAM_OPEN[wAdapterIndex][i].h.wError =
				HPI_ERROR_INVALID_OBJ;
			aRESP_HPI_ISTREAM_OPEN[wAdapterIndex][i].h.wError =
				HPI_ERROR_INVALID_OBJ;
		}
	}
}

static uint16_t HPIMSGX_Init(
	struct hpi_message *phm,
	// HPI_SUBSYS_CREATE_ADAPTER structure with
	// resource list or NULL=find all
	struct hpi_response *phr
	// response from HPI_ADAPTER_GET_INFO
)
{
	HPI_HandlerFunc *entry_point_func;
	struct hpi_response hr;


	// Init response here so we can pass in previous adapter list
	HPI_InitResponse(&hr, phm->wObject, phm->wFunction,
		HPI_ERROR_INVALID_OBJ);

	entry_point_func =
		HPI_LookupEntryPointFunction(phm->u.s.resource.r.pci);

	if (entry_point_func) {
		HPI_DEBUG_MESSAGE(DEBUG, phm);
		entry_point_func(phm, &hr);
	} else {
		phr->wError = HPI_ERROR_PROCESSING_MESSAGE;
		return phr->wError;
	}
	if (hr.wError == 0) {
		/* the adapter was created successfully
		   save the mapping for future use */
		hpi_entry_points[hr.u.s.wAdapterIndex] = entry_point_func;
		// prepare adapter (pre-open streams etc.)
		HPI_DEBUG_LOG0(DEBUG,
			"HPI_SUBSYS_CREATE_ADAPTER successful,"
			" preparing adapter\n");
		AdapterPrepare(hr.u.s.wAdapterIndex);
	}
	memcpy(phr, &hr, hr.wSize);
	return phr->wError;
}

static void HPIMSGX_Cleanup(
	uint16_t wAdapterIndex,
	void *hOwner
)
{
	int i, wAdapter, wAdapterLimit;

	if (!hOwner)
		return;

	if (wAdapterIndex == HPIMSGX_ALLADAPTERS) {
		wAdapter = 0;
		wAdapterLimit = HPI_MAX_ADAPTERS;
	} else {
		wAdapter = wAdapterIndex;
		wAdapterLimit = wAdapter + 1;
	}

	for (; wAdapter < wAdapterLimit; wAdapter++) {
		//      printk(KERN_INFO "Cleanup adapter #%d\n",wAdapter);
		for (i = 0; i < HPI_MAX_STREAMS; i++) {
			if (hOwner == aOStreamUserOpen[wAdapter][i].hOwner) {
				struct hpi_message hm;
				struct hpi_response hr;

				HPI_DEBUG_LOG2(DEBUG,
					"Close adapter %d ostream %d\n",
					wAdapter, i);

				HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_OSTREAM,
					HPI_OSTREAM_RESET);
				hm.wAdapterIndex = (uint16_t)wAdapter;
				hm.wObjIndex = (uint16_t)i;
				HwEntryPoint(&hm, &hr);

				hm.wFunction = HPI_OSTREAM_HOSTBUFFER_FREE;
				HwEntryPoint(&hm, &hr);

				hm.wFunction = HPI_OSTREAM_GROUP_RESET;
				HwEntryPoint(&hm, &hr);

				aOStreamUserOpen[wAdapter][i].nOpenFlag = 0;
				aOStreamUserOpen[wAdapter][i].hOwner = NULL;
			}
			if (hOwner == aIStreamUserOpen[wAdapter][i].hOwner) {
				struct hpi_message hm;
				struct hpi_response hr;

				HPI_DEBUG_LOG2(DEBUG,
					"Close adapter %d istream %d\n",
					wAdapter, i);

				HPI_InitMessageResponse(&hm, &hr, HPI_OBJ_ISTREAM,
					HPI_ISTREAM_RESET);
				hm.wAdapterIndex = (uint16_t)wAdapter;
				hm.wObjIndex = (uint16_t)i;
				HwEntryPoint(&hm, &hr);

				hm.wFunction = HPI_ISTREAM_HOSTBUFFER_FREE;
				HwEntryPoint(&hm, &hr);

				hm.wFunction = HPI_ISTREAM_GROUP_RESET;
				HwEntryPoint(&hm, &hr);

				aIStreamUserOpen[wAdapter][i].nOpenFlag = 0;
				aIStreamUserOpen[wAdapter][i].hOwner = NULL;
			}
		}
	}
}

