/* radio-asihpi.c
 *
 * A device driver for AudioScience radio cards.
 *
 * Copyright 2004 Fred Gleason <fredg@salemradiolabs.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <hpi.h>

#include <linux/version.h>
#include <linux/autoconf.h>
#include <linux/module.h>
#include <linux/videodev2.h>

/* V4L helper functions moved to new header file since 2.6.18.1 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
#include <media/v4l2-common.h>
#include <media/v4l2-ioctl.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
#define NO_V4L2_FOPS
#endif

/*
 * Globals
 */
static void *ss;

struct radio_asihpi_tuner {
	hpi_handle_t h;
	int bands;		/* 0 = TV/FM, 1 = AM/FM */
	int index;
	struct video_device videodev;
};

static struct radio_asihpi_tuner radio_asihpi_tuner[HPI_MAX_ADAPTERS][8];

static int radio_asihpi_getband(struct radio_asihpi_tuner *t)
{
	u16 hpi_band = 0;

	if (!t->h)
		return -1;

	if (HPI_Tuner_GetBand(ss, t->h, &hpi_band) != 0) {
		return -1;
	}

	switch (t->bands) {
	case 0:		/* FM/TV */
		switch (hpi_band) {
		case HPI_TUNER_BAND_FM:
			return 0;
		case HPI_TUNER_BAND_FM_STEREO:
			return 1;
		case HPI_TUNER_BAND_TV:
			return 2;
		}
		break;
	case 1:		/* FM/AM */
		switch (hpi_band) {
		case HPI_TUNER_BAND_FM:
			return 0;
		case HPI_TUNER_BAND_FM_STEREO:
			return 1;
		case HPI_TUNER_BAND_AM:
			return 2;
		}
		break;
	}
	return -1;
}

static int radio_asihpi_setband(struct radio_asihpi_tuner *t, unsigned band)
{
	u16 hpi_band = 0;

	if (!t->h)
		return -1;

	switch (t->bands) {
	case 0:		/* FM/TV */
		switch (band) {
		case 0:
			hpi_band = HPI_TUNER_BAND_FM;
			break;

		case 1:
			hpi_band = HPI_TUNER_BAND_FM_STEREO;
			break;

		case 2:
			hpi_band = HPI_TUNER_BAND_TV;
			break;
		}
		break;

	case 1:		/* FM/AM */
		switch (band) {
		case 0:
			hpi_band = HPI_TUNER_BAND_FM;
			break;

		case 1:
			hpi_band = HPI_TUNER_BAND_FM_STEREO;
			break;

		case 2:
			hpi_band = HPI_TUNER_BAND_AM;
			break;
		}
		break;
	}

	if (HPI_Tuner_SetBand(ss, t->h, hpi_band) != 0) {
		return -1;
	}
	return 0;
}

static int radio_asihpi_getsignal(struct radio_asihpi_tuner *t)
{
	short hpi_signal = 0;
	if (!t->h)
		return -1;

	if (HPI_Tuner_GetRFLevel(ss, t->h, &hpi_signal) != 0)
		return -1;

    	printk(KERN_INFO "radio_asihpi signal %d\n", hpi_signal * 5);
	return hpi_signal * 5;
}

static int receiving_stereo(struct radio_asihpi_tuner *t)
{
	u16 s,sm;

	if (!t->h)
		return -1;

        if (HPI_Tuner_GetStatus(ss, t->h, &sm, &s) != 0)
		return 0;

    	printk(KERN_INFO "radio_asihpi stereo %d\n", s & HPI_TUNER_FM_STEREO);

	return s & HPI_TUNER_FM_STEREO;
}

static int radio_asihpi_getfreq(struct radio_asihpi_tuner *t)
{
	u16 hpi_band = 0;
	u32 hpi_freq = 0;

	if (!t->h)
		return -1;

	if (HPI_Tuner_GetFrequency(ss, t->h, &hpi_freq) != 0) {
		return -1;
	}
	if (HPI_Tuner_GetBand(ss, t->h, &hpi_band) != 0) {
		return -1;
	}
	switch (hpi_band) {
	case HPI_TUNER_BAND_FM:
	case HPI_TUNER_BAND_FM_STEREO:
	case HPI_TUNER_BAND_TV:
	case HPI_TUNER_BAND_AM:
		return 16 * hpi_freq;
	}
	return -1;
}

static int radio_asihpi_setfreq(struct radio_asihpi_tuner *t, unsigned freq)
{
	u32 hpi_freq = 0;

	if (!t->h)
		return -1;


	switch (t->bands) {
	case 0:		/* FM/TV */
		switch (radio_asihpi_getband(t)) {
		case 0:	/* FM */
		case 1:	/* FM Stereo */
			if ((freq < 1216000) || (freq > 1728000)) {
				return -1;
			}
			break;

		case 2:	/* TV */
			if ((freq < 884000) || (freq > 12820000)) {
				return -1;
			}
			break;
		}
		break;

	case 1:		/* FM/AM */
		switch (radio_asihpi_getband(t)) {
		case 0:	/* FM */
		case 1:	/* FM Stereo */
			if ((freq < 1216000) || (freq > 1728000)) {
				return -1;
			}
			break;

		case 2:	/* AM */
			if ((freq < 8320) || (freq > 28000)) {
				return -1;
			}
			break;
		}
		break;
	}
	hpi_freq = freq / 16;
	if (HPI_Tuner_SetFrequency(ss, t->h, hpi_freq) != 0) {
		return -1;
	}
	return 0;
}

#ifdef NO_V4L2_FOPS
static int radio_asihpi_open(struct inode *inode, struct file *file)
#else
static int radio_asihpi_open(struct file *file)
#endif
{
	int card;
	int tuner;
	int found = 0;
	struct radio_asihpi_tuner *t;

	for (card = 0; card < HPI_MAX_ADAPTERS; card++) {
		for (tuner = 0; tuner < 8; tuner++) {
			t = &radio_asihpi_tuner[card][tuner];
			if (t->h &&
			    (t->videodev.minor ==
			     MINOR(file->f_dentry->d_inode->i_rdev))) {
				found = 1;
				break;
			}
		}
		if (found)
			break;
	}
	if (!found) {
		return -EINVAL;	// Is this really the approriate return here?
	}

	file->private_data = t;
	return 0;
}

#ifdef NO_V4L2_FOPS
static int radio_asihpi_release(struct inode *inode, struct file *file)
#else
static int radio_asihpi_release(struct file *file)
#endif
{
    	printk(KERN_INFO "radio_asihpi_release\n");
	return 0;
}

/*** V4L2 ioctl callbacks ***/
static int vidioc_querycap (struct file *file, void  *priv,
					struct v4l2_capability *cap)
{
	strcpy(cap->driver, "radio-asihpi");
	strcpy(cap->card, "asihpi");
	cap->version = 0;
	cap->capabilities = 	V4L2_CAP_TUNER |
				V4L2_CAP_RADIO;
	//			V4L2_CAP_AUDIO |
	// V4L2_CAP_RDS_CAPTURE

	printk(KERN_INFO "vidioc_querycap\n");
	return 0;
}

/*
struct v4l2_tuner
{
	__u32                   index;
	__u8			name[32];
	enum v4l2_tuner_type    type;
	__u32			capability;
	__u32			rangelow;
	__u32			rangehigh;
	__u32			rxsubchans;
	__u32			audmode;
	__s32			signal;
	__s32			afc;
	__u32			reserved[4];
};
*/

int vidioc_g_tuner(struct file *file, void *fh,	struct v4l2_tuner *a)
{
	struct radio_asihpi_tuner *t = file->private_data;

	a->type = V4L2_TUNER_RADIO;
	a->signal = radio_asihpi_getsignal(t);
	a->capability = V4L2_TUNER_CAP_LOW | V4L2_TUNER_CAP_STEREO;
	if (receiving_stereo(t))
		a->rxsubchans = V4L2_TUNER_SUB_STEREO;
	else
		a->rxsubchans = V4L2_TUNER_SUB_MONO;

	switch (t->bands) {
	case 0:	/* FM/TV */
		switch (a->index) {
		case 0:
			strcpy(a->name, "FM");
			a->rangelow = 1216000;	// 76.0 MHz
			a->rangehigh = 1728000;	// 108.0 MHz
			break;
		case 1:
			strcpy(a->name, "FM Stereo");
			a->rangelow = 1216000;	// 76.0 MHz
			a->rangehigh = 1728000;	// 108.0 MHz
			break;
		case 2:
			strcpy(a->name, "TV");
			a->rangelow = 884000;	// 55.25 MHz
			a->rangehigh = 12820000;	// 801.25 MHz
			break;
		default:
			return -EINVAL;
		}
		break;
	case 1:	/* FM/AM */
		switch (a->index) {
		case 0:
			strcpy(a->name, "FM");
			a->rangelow = 1216000;	// 76.0 MHz
			a->rangehigh = 1728000;	// 108.0 MHz
			break;
		case 1:
			strcpy(a->name, "FM Stereo");
			a->rangelow = 1216000;	// 76.0 MHz
			a->rangehigh = 1728000;	// 108.0 MHz
			break;
		case 2:
			strcpy(a->name, "AM");
			a->rangelow = 8320;	// 520 kHz
			a->rangehigh = 28000;	// 1750 kHz
			break;
		default:
			return -EINVAL;
		}
		break;

	default:
		return -EINVAL;
	}

	return 0;
}

int vidioc_s_tuner(struct file *file, void *fh, struct v4l2_tuner *a)
{
	struct radio_asihpi_tuner *t = file->private_data;

	if (radio_asihpi_setband(t, a->index) < 0)
		return -EINVAL;

	return 0;
}

int vidioc_g_frequency(struct file *file, void *fh, struct v4l2_frequency *a)
{
	struct radio_asihpi_tuner *t = file->private_data;
	int freq;

	if ((freq = radio_asihpi_getfreq(t)) < 0) {
		return -EINVAL;
	}

	a->frequency = freq;
	return 0;
}

int vidioc_s_frequency(struct file *file, void *fh, struct v4l2_frequency *a)
{
	struct radio_asihpi_tuner *t = file->private_data;

	if (radio_asihpi_setfreq(t, a->frequency) < 0) {
		return -EINVAL;
	}
	return 0;
}

#ifdef NO_V4L2_FOPS
static const struct file_operations radio_asihpi_fops =
#else
static const struct v4l2_file_operations radio_asihpi_fops =
#endif
{
	.owner = THIS_MODULE,
	.open = radio_asihpi_open,
	.release = radio_asihpi_release,
	.ioctl = video_ioctl2,
};

static const struct v4l2_ioctl_ops radio_asihpi_ioctl_ops = {
	.vidioc_querycap = vidioc_querycap,
	.vidioc_g_tuner = vidioc_g_tuner,
	.vidioc_s_tuner = vidioc_s_tuner,
	.vidioc_g_frequency = vidioc_g_frequency,
	.vidioc_s_frequency = vidioc_s_frequency,
};

static void radio_asihpi_device_release(struct video_device *vfd)
{
	printk(KERN_INFO "radio asihpi device release\n");
}

int init_module(void)
{
	int i, j, k;
	int found = 0;
	int num_adap;
	hpi_handle_t h_mixer;
	hpi_handle_t h_control;
	u16 hpi_band;

	if (HPI_SubSysGetNumAdapters(ss, &num_adap) != 0 || num_adap == 0)
		return -ENODEV;

	memset(radio_asihpi_tuner, 0,sizeof(radio_asihpi_tuner));

	for (i = 0; i < num_adap; i++) {
		u32 idx;
		u16 type;

		if (HPI_SubSysGetAdapter(ss, i, &idx, &type))
			continue;

		if ((type & 0xF000) != 0x8000)
			continue;

		if (HPI_MixerOpen(ss, idx, &h_mixer) != 0)
			continue;

		for (j = 0; j < 8; j++) {
			struct radio_asihpi_tuner *t;
			t = &radio_asihpi_tuner[idx][j];

			if (HPI_MixerGetControl
			    (ss, h_mixer,
			     HPI_SOURCENODE_TUNER, j, 0, 0,
			     HPI_CONTROL_TUNER, &h_control) != 0)
				continue;

			k = 0;
			while (HPI_Tuner_QueryBand
			       (ss, h_control,
				k,  &hpi_band) == 0) {
				if (hpi_band == HPI_TUNER_BAND_AM) {
					t->bands = 1;
				}
				k++;
			}
			sprintf(t->videodev.name,
				"AudioScience %4x[%d] Tuner %d",
				type, idx + 1, j + 1);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 27)
			t->videodev.owner = THIS_MODULE;
			t->videodev.type =  VFL_TYPE_RADIO;
#else
			t->videodev.vfl_type =  VFL_TYPE_RADIO;
#endif

			t->videodev.fops = &radio_asihpi_fops;
			t->videodev.ioctl_ops = &radio_asihpi_ioctl_ops;
			t->videodev.release = radio_asihpi_device_release;

			t->videodev.debug = 1;

			if (video_register_device
			    (&t->videodev, VFL_TYPE_RADIO, -1) == 0) {
				t->h = h_control;
				t->index = 8 * idx + j + 1;
				found = 1;

                                printk(KERN_INFO "Register adap %d tuner %d\n",idx,j);

			}
		}
		if (HPI_MixerClose(ss, h_mixer)) {
			/* deliberately ignore any error */
		}
	}
	if (!found)
		return -ENODEV;

        printk(KERN_INFO "radio-asihpi init_module OK\n");
	return 0;
}

void cleanup_module(void)
{
	int idx;
	int j;

	for (idx = 0; idx < HPI_MAX_ADAPTERS; idx++) {
		for (j = 0; j < 8; j++) {
			if (radio_asihpi_tuner[idx][j].h) {
                                printk(KERN_INFO "UnRegister adap %d tuner %d\n",idx,j);
				video_unregister_device(&radio_asihpi_tuner[idx]
							[j].videodev);
			}
		}
	}
}

MODULE_AUTHOR("Fred Gleason <fredg@salemradiolabs.com>");
MODULE_DESCRIPTION("A V4L driver for AudioScience radio cards");
MODULE_LICENSE("GPL");
